@@ -40,7 +40,7 @@
 //! let expected = 1;
 //! assert!(expected == ans);
 //! ```
-//! 
+//!
 use std::cmp::Ordering;
 
 #[derive(Debug, Clone)]
@@ -92,6 +92,7 @@ where
     level: usize,
     cnt: usize,
     s: Vec<Vec<RankInfo<T>>>,
+    cached_s_m: Option<Vec<RankInfo<T>>>,
 }
 
 impl<T> FixedSizeEpsilonSummary<T>
@@ -123,10 +124,12 @@ where
             level: number_of_levels,
             cnt: 0,
             s,
+            cached_s_m: None,
         }
     }
 
     pub fn update(&mut self, e: T) {
+        self.cached_s_m = None;
         let rank_info = RankInfo::new(e, 0, 0);
         self.s[0].push(rank_info);
 
@@ -142,7 +145,7 @@ where
         }
 
         let compressed_size = self.b / 2;
-        let mut s_c = compress(&self.s[0], compressed_size, self.epsilon);
+        let mut s_c = compress(self.s[0].clone(), compressed_size, self.epsilon);
 
         self.s[0].clear();
         for k in 1..self.level {
@@ -151,27 +154,31 @@ where
                 break;
             } else {
                 let t = merge(s_c, &self.s[k]);
-                s_c = compress(&t, compressed_size, self.epsilon);
+                s_c = compress(t, compressed_size, self.epsilon);
                 self.s[k].clear();
             }
         }
     }
 
     pub fn query(&mut self, r: f64) -> T {
-        self.s[0].sort();
-        for (i, r) in self.s[0].iter_mut().enumerate() {
-            r.rmin = i as i64;
-            r.rmax = i as i64;
-        }
+        if self.cached_s_m.is_none() {
+            self.s[0].sort();
+            for (i, r) in self.s[0].iter_mut().enumerate() {
+                r.rmin = i as i64;
+                r.rmax = i as i64;
+            }
 
-        let mut s_m = self.s[0].clone();
-        for i in 1..self.level {
-            s_m = merge(s_m, &self.s[i])
+            let mut s_m = self.s[0].clone();
+            for i in 1..self.level {
+                s_m = merge(s_m, &self.s[i])
+            }
+
+            self.cached_s_m = Some(s_m);
         }
 
         let rank: i64 = ((self.cnt as f64) * r).floor() as i64;
         let epsilon_n: i64 = ((self.cnt as f64) * self.epsilon).floor() as i64;
-        let e = find_idx(&s_m, rank, epsilon_n).unwrap();
+        let e = find_idx(&self.cached_s_m.as_ref().unwrap(), rank, epsilon_n).unwrap();
         return e;
     }
 
@@ -187,7 +194,7 @@ where
             s_m = merge(s_m, &self.s[i])
         }
 
-        compress(&s_m, self.b, epsilon)
+        compress(s_m, self.b, epsilon)
     }
 
     fn finalize(&mut self, epsilon: f64) {
@@ -273,13 +280,11 @@ fn merge<T: Clone + Ord>(s_a: Vec<RankInfo<T>>, s_b: &[RankInfo<T>]) -> Vec<Rank
     s_m
 }
 
-fn compress<T: Clone>(s0: &[RankInfo<T>], block_size: usize, epsilon: f64) -> Vec<RankInfo<T>> {
-    let mut s_c = Vec::with_capacity(s0.len());
-
+fn compress<T: Clone>(mut s0: Vec<RankInfo<T>>, block_size: usize, epsilon: f64) -> Vec<RankInfo<T>> {
     let mut s0_range = 0;
     let mut e: f64 = 0.0;
 
-    for r in s0 {
+    for r in &s0 {
         if s0_range < r.rmax {
             s0_range = r.rmax;
         }
@@ -294,10 +299,12 @@ fn compress<T: Clone>(s0: &[RankInfo<T>], block_size: usize, epsilon: f64) -> Ve
 
     let mut i = 0;
     let mut j = 0;
-    while i <= block_size && j < s0.len() {
+    let mut k = 0;
+    let n = s0.len();
+    while i <= block_size && j < n {
         let r = ((i as f64) * (s0_range as f64) / (block_size as f64)).floor() as i64;
 
-        while j < s0.len() {
+        while j < n {
             if s0[j].rmax >= r {
                 break;
             }
@@ -305,38 +312,34 @@ fn compress<T: Clone>(s0: &[RankInfo<T>], block_size: usize, epsilon: f64) -> Ve
             j += 1;
         }
 
-        assert!(j < s0.len(), "unable to find the summary with precision given.");
-        s_c.push(s0[j].clone());
+        assert!(j < n, "unable to find the summary with precision given.");
+        s0[k] = s0[j].clone();
+        k += 1;
         j += 1;
         i += 1;
     }
 
-    s_c
-}
-
-#[inline]
-fn boundary_val(i: u32, epsilon: f64) -> usize {
-    let n = usize::pow(2, i);
-    (((n - 1) as f64) / epsilon).floor() as usize
+    s0.truncate(k);
+    s0
 }
 
 #[inline]
 #[allow(clippy::many_single_char_names)]
 #[allow(clippy::comparison_chain)]
-fn is_boundary(x: usize, epsilon: f64) -> Option<u32> {
-    let mut l = 0u32;
-    let mut r = 31u32;
+fn is_boundary(x: usize, boundaries: &[usize; 32]) -> Option<usize> {
+    let mut l = 0;
+    let mut r = 31;
 
     while l < r {
         let m = l + (r - l) / 2;
-        if boundary_val(m, epsilon) < x {
+        if boundaries[m] < x {
             l = m + 1;
         } else {
             r = m;
         }
     }
 
-    if l <= 31u32 && boundary_val(l, epsilon) == x {
+    if l < 31 && boundaries[l] == x {
         return Some(l);
     }
 
@@ -350,6 +353,8 @@ fn find_idx<T: Clone + Ord>(s_m: &[RankInfo<T>], rank: i64, epsilon_n: i64) -> O
         let m = l + (r - l) / 2;
         if s_m[m].rmin < rank {
             l = m + 1;
+        } else if s_m[m].rmax > rank {
+            r = m;
         } else {
             r = m;
         }
@@ -374,6 +379,8 @@ where
     cnt: usize,
     s: Vec<FixedSizeEpsilonSummary<T>>,
     s_c: FixedSizeEpsilonSummary<T>,
+    boundaries: [usize; 32],
+    cached_s_m: Option<Vec<RankInfo<T>>>,
 }
 
 impl<T> UnboundEpsilonSummary<T>
@@ -385,40 +392,54 @@ where
 
         let n = (1.0_f64 / epsilon).floor() as usize;
         let s_c = FixedSizeEpsilonSummary::new(n, epsilon / 2.0);
+        let mut boundaries: [usize; 32] = [0; 32];
+        for i in 0..32usize {
+            let boundary = (((usize::pow(2, i as u32) - 1) as f64) / epsilon).floor() as usize;
+            boundaries[i] = boundary
+        }
+
         UnboundEpsilonSummary {
             epsilon,
             cnt: 0,
             s,
             s_c,
+            boundaries,
+            cached_s_m: None,
         }
     }
 
     pub fn update(&mut self, e: T) {
+        self.cached_s_m = None;
         self.s_c.update(e);
 
-        if let Some(x) = is_boundary(self.cnt + 1, self.epsilon) {
+        if let Some(x) = is_boundary(self.cnt + 1, &self.boundaries) {
             self.s_c.finalize(self.epsilon / 2.0);
-            self.s.push(self.s_c.clone());
 
-            let upper_bound = (((usize::pow(2, x + x) - 1) as f64) / self.epsilon).floor() as usize;
+            let upper_bound = self.boundaries[x + x];
             let n = upper_bound - self.cnt - 1;
-            let summary = FixedSizeEpsilonSummary::new(n, self.epsilon / 2.0);
-            self.s_c = summary;
+            let mut summary = FixedSizeEpsilonSummary::new(n, self.epsilon / 2.0);
+            std::mem::swap(&mut self.s_c, &mut summary);
+
+            self.s.push(summary);
         }
         self.cnt += 1;
     }
 
     pub fn query(&mut self, r: f64) -> T {
-        let mut s_m = self.s_c.calc_s_m(self.epsilon / 2.0);
-        for i in 0..self.s.len() {
-            for j in 0..self.s[i].s.len() {
-                s_m = merge(s_m, &self.s[i].s[j])
+        if self.cached_s_m.is_none() {
+            let mut s_m = self.s_c.calc_s_m(self.epsilon / 2.0);
+            for i in 0..self.s.len() {
+                for j in 0..self.s[i].s.len() {
+                    s_m = merge(s_m, &self.s[i].s[j])
+                }
             }
+
+            self.cached_s_m = Some(s_m);
         }
 
         let rank: i64 = ((self.cnt as f64) * r).floor() as i64;
         let epsilon_n: i64 = ((self.cnt as f64) * self.epsilon).floor() as i64;
-        let e = find_idx(&s_m, rank, epsilon_n).unwrap();
+        let e = find_idx(&self.cached_s_m.as_ref().unwrap(), rank, epsilon_n).unwrap();
         return e;
     }
 
@@ -463,7 +484,7 @@ mod tests {
         assert_eq!(merged_rmaxs, vec![1, 3, 6, 8, 11, 13, 15, 16]);
 
         let epsilon: f64 = 0.2;
-        let compressed = compress(&merged, 4, epsilon);
+        let compressed = compress(merged, 4, epsilon);
         let compressed_vals: Vec<i32> = compressed.iter().map(|x| x.val).collect();
         assert_eq!(compressed_vals, vec![1, 4, 7, 12, 17]);
     }
@@ -500,6 +521,94 @@ mod tests {
         }
     }
 
+    #[test]
+    fn test_normal_distribution_generated_seq_on_fixed_summary() {
+        let n = 1000000;
+        let epsilon: f64 = 0.01;
+        let mut s = FixedSizeEpsilonSummary::new(n, epsilon);
+
+        let dn = rand_distr::Normal::new(0.5f64, 0.2f64).unwrap();
+        let mut normal_rng = rand_pcg::Pcg64::new(0xcafef00dd15ea5e5, 0xa02bdbf7bb3c0a7ac28fa16a64abf96);
+
+        let startmem = stats_alloc::Region::new(&GLOBAL);
+        let mut records = Vec::with_capacity(n);
+        for _ in 0..n {
+            let x = dn.sample(&mut normal_rng);
+            records.push(unsafe { ordered_float::NotNan::new_unchecked(x) });
+        }
+        let mem = startmem.change();
+        let bytes_change = mem.bytes_allocated as isize - mem.bytes_deallocated as isize + mem.bytes_reallocated;
+
+        println!("{}k", bytes_change / 1024);
+
+        records.sort_by(|a, b| a.partial_cmp(b).unwrap());
+
+        let startmem = stats_alloc::Region::new(&GLOBAL);
+        for i in 0..n {
+            s.update(records[i]);
+        }
+        let mem = startmem.change();
+        let bytes_change = mem.bytes_allocated as isize - mem.bytes_deallocated as isize + mem.bytes_reallocated;
+
+        println!("{}k", bytes_change / 1024);
+
+        let quantile_estimated = s.query(0.5);
+        assert!((quantile_estimated - records[n / 2]).abs() < 0.01);
+
+        let quantile_estimated = s.query(0.0);
+        assert!((quantile_estimated - records[0]).abs() < 0.1);
+
+        let quantile_estimated = s.query(0.99);
+        assert!((quantile_estimated - records[n * 99 / 100]).abs() < 0.01);
+
+        let quantile_estimated = s.query(1.0);
+        assert!((quantile_estimated - records[n - 1]).abs() < 0.01);
+    }
+
+    #[test]
+    fn test_pareto_distribution_generated_seq_on_fixed_summary() {
+        let n = 1000000;
+        let epsilon: f64 = 0.001;
+        let mut s = FixedSizeEpsilonSummary::new(n, epsilon);
+
+        let dn = rand_distr::Pareto::new(5f64, 10f64).unwrap();
+        let mut normal_rng = rand_pcg::Pcg64::new(0xcafef00dd15ea5e5, 0xa02bdbf7bb3c0a7ac28fa16a64abf96);
+
+        let startmem = stats_alloc::Region::new(&GLOBAL);
+        let mut records = Vec::with_capacity(n);
+        for _ in 0..n {
+            let x = dn.sample(&mut normal_rng);
+            records.push(unsafe { ordered_float::NotNan::new_unchecked(x) });
+        }
+        let mem = startmem.change();
+        let bytes_change = mem.bytes_allocated as isize - mem.bytes_deallocated as isize + mem.bytes_reallocated;
+
+        println!("{}k", bytes_change / 1024);
+
+        records.sort_by(|a, b| a.partial_cmp(b).unwrap());
+
+        let startmem = stats_alloc::Region::new(&GLOBAL);
+        for i in 0..n {
+            s.update(records[i]);
+        }
+        let mem = startmem.change();
+        let bytes_change = mem.bytes_allocated as isize - mem.bytes_deallocated as isize + mem.bytes_reallocated;
+
+        println!("{}k", bytes_change / 1024);
+
+        let quantile_estimated = s.query(0.5);
+        assert!((quantile_estimated - records[n / 2]).abs() < 0.01);
+
+        let quantile_estimated = s.query(0.0);
+        assert!((quantile_estimated - records[0]).abs() < 0.01);
+
+        let quantile_estimated = s.query(0.99);
+        assert!((quantile_estimated - records[n * 99 / 100]).abs() < 0.01);
+
+        let quantile_estimated = s.query(1.0);
+        assert!((quantile_estimated - records[n - 1]).abs() < 0.01);
+    }
+
     #[test]
     fn test_normal_distribution_generated_seq_on_unbound_summary() {
         let n = 1000000;
